import { Buffer } from "buffer"
import { createDecipheriv } from "../adapters/decipher"
import type { IAdapter } from "../adapters"
import { createEventEmitter } from "../ee"
import { HMACAssertionError } from "../errors"
import type { i18n } from "../i18n"
import type { ItemDetails, Overview, Profile } from "../types"
import { setIfAbsent } from "../util"
import type { EncryptedItem } from "./item"

/** Encryption and MAC */
export interface Cipher {
  /** Encryption key */
  key: Buffer
  /** HMAC key */
  hmac: Buffer
}

export class Crypto implements IDisposable {
  #disposables: IDisposable[] = []
  #locked = true

  #master!: Cipher
  #overview!: Cipher

  private subtle: SubtleCrypto
  private hmacSHA256: IAdapter["hmacSHA256"]

  readonly onLock = createEventEmitter<void>()

  constructor(private readonly i18n: i18n, adapter: IAdapter) {
    this.subtle = adapter.subtle
    this.hmacSHA256 = adapter.hmacSHA256
  }

  async unlock(profile: Profile, masterPassword: string) {
    const encoder = new TextEncoder()
    const key = await this.subtle.importKey(
      "raw",
      encoder.encode(masterPassword),
      { name: "PBKDF2" },
      false,
      ["deriveBits"]
    )
    const derivedKey = await this.subtle.deriveBits(
      {
        name: "PBKDF2",
        salt: Buffer.from(profile.salt, "base64"),
        iterations: profile.iterations,
        hash: {
          name: "SHA-512",
        },
      },
      key,
      64 << 3
    )

    const cipher = splitPlainText(Buffer.from(derivedKey))

    // Derive master key and overview keys
    this.#master = await this.decryptKeys(profile.masterKey, cipher)
    this.#overview = await this.decryptKeys(profile.overviewKey, cipher)
    this.#locked = false
  }

  get locked() {
    return this.#locked
  }

  lock() {
    this.#locked = true
    this.#master = null!
    this.#overview = null!
    this.#disposables.forEach(fn => fn.dispose())
    this.onLock()
  }

  dispose() {
    this.lock()
  }

  assertUnlocked() {
    if (this.#locked) {
      throw new Error(this.i18n.error.vaultIsLocked)
    }
  }

  #createCache = <K, V, K2 = K>(
    deriveArg: (value: K) => K2,
    implementation: (value: K2) => V
  ) => {
    const map = new Map<K2, V>()
    this.#disposables.push({
      dispose: () => map.clear(),
    })
    return (data: K) => setIfAbsent(map, deriveArg(data), implementation)
  }

  #createWeakCache = <K extends object, V>(implementation: (value: K) => V) => {
    let map = new WeakMap<K, V>()
    this.#disposables.push({
      dispose() {
        map = new WeakMap()
      },
    })
    return (data: K) => setIfAbsent(map, data, implementation)
  }

  decryptItemDetails = this.#createWeakCache((item: EncryptedItem) => {
    const cipher = this.deriveConcreteKey(item)
    const detail = this.decryptOPData(Buffer.from(item.d, "base64"), cipher)
    return JSON.parse(detail.toString("utf-8")) as ItemDetails
  })

  decryptItemOverview = this.#createCache(
    (item: EncryptedItem) => item.o,
    (o: string) => {
      const overview = this.decryptOPData(Buffer.from(o, "base64"), this.#overview)
      return JSON.parse(overview.toString("utf8")) as Overview
    }
  )

  deriveConcreteKey = this.#createCache(
    (data: { k: string }) => data.k,
    ($k: string) => {
      const k = Buffer.from($k, "base64")
      const data = k.slice(0, -32)
      this.assertHMac(data, this.#master.hmac, k.slice(-32))
      const derivedKey = decryptData(this.#master.key, data.slice(0, 16), data.slice(16))
      return splitPlainText(derivedKey)
    }
  )

  assertHMac(data: Buffer, key: Buffer, expected: Buffer) {
    const actual = this.hmacSHA256(key, data)
    if (!actual.equals(expected)) {
      throw new HMACAssertionError()
    }
  }

  decryptOPData(cipherText: Buffer, cipher: Cipher) {
    const key = cipherText.slice(0, -32)
    this.assertHMac(key, cipher.hmac, cipherText.slice(-32))

    const plaintext = decryptData(cipher.key, key.slice(16, 32), key.slice(32))
    const size = readUint16(key.slice(8, 16))
    return plaintext.slice(-size)
  }

  async decryptKeys(encryptedKey: string, derived: Cipher) {
    const buffer = Buffer.from(encryptedKey, "base64")
    const base = this.decryptOPData(buffer, derived)
    const digest = await this.subtle.digest("SHA-512", base)
    return splitPlainText(Buffer.from(digest))
  }

  get overview() {
    return this.#overview
  }
}

export const splitPlainText = (derivedKey: Buffer): Cipher => ({
  key: derivedKey.slice(0, 32),
  hmac: derivedKey.slice(32, 64),
})

function decryptData(key: Buffer, iv: Buffer, data: Buffer) {
  return createDecipheriv("aes-256-cbc", key, iv).setAutoPadding(false).update(data)
}

function readUint16({ buffer, byteOffset, length }: Buffer) {
  return new DataView(buffer, byteOffset, length).getUint16(0, true)
}
